/*
 *   Java-Gnome
 *
 *   Copyright (C) 1998 Olivier Gutknecht <gutkneco@lirmm.fr>
 *
 *   The Java-Gnome bindings library is free software distributed under
 *   the terms of the GNU Library General Public License version 2 
 *
 ***********************************************************************/


/** A class defining a Java type in the context of type conversion
for JNI prototype and glue code generation.

BUGS: enum handling is not implemented yet.

@author Olivier Gutknecht
*/


import java.util.*;

public class TypeDefinition {
    public final static int UNKNOWN = 0;
    public final static int BOOLEAN = 1;
    public final static int BYTE = 2;
    public final static int CHAR = 3;
    public final static int DOUBLE = 4;
    public final static int FLOAT = 5;
    public final static int INTEGER = 6;
    public final static int LONG = 7;
    public final static int SHORT = 8;
    public final static int VOID = 9;
    public final static int STRING = 10;
    public final static int OBJECT = 11;
    public final static int ENUM = 12;
    public final static int TIME_T = 13;
    public final static int POINTER = 14;
    public final static int GLIST = 15;
    
    /* A java.lang.String is an object, but we identify separately
       because it corresponds to a special JNI type */
    
    int type = 0;
    boolean array = false;
    String value = null;
    
    public static String checkAliases(String s) {
	if (s.equals("file-descriptor"))
	    return "java.io.FileDescriptor";
	if (s.equals("callback"))
	    return "Callback";
	if (s.equals("full-callback"))
	    return "Callback";
	if (s.equals("type"))
	    return "int";
	if (s.equals("GdkInputCondition"))
	    return "int";
	return s;
    }
    
    public static TypeDefinition convertFromScheme(String s) {
	TypeDefinition td = new TypeDefinition();
	if (s.endsWith("*")) {
	    td.array = true;
	    s = s.substring(0, s.length() - 1);
	}
	if (s.equals("none"))
	    td.type = VOID;
	else if (s.equals("bool"))
	    td.type = BOOLEAN;
	else if (s.equals("float"))
	    //td.type = FLOAT;
	    // FIXME: is that correct on C side ? 
	    td.type = DOUBLE;
	else if (s.equals("double"))
	    td.type = DOUBLE;
	else if (s.equals("uint") || s.equals("int"))
	    td.type = INTEGER;
	else if (s.equals("long") || s.equals("ulong"))
	    td.type = LONG;
	else if (s.equals("string") || s.equals("static_string"))
	    td.type = STRING;
	else if (s.equals("time_t"))
	    td.type = TIME_T;
	else if (s.equals("gpointer"))
	    td.type = POINTER;
	else if (s.equals("GList"))
	    td.type = GLIST;
	else {
	    td.type = OBJECT;
	    td.value = checkAliases(s);
	}
	return td;
    }
    
    public boolean equals(Object o) {
	if (o instanceof TypeDefinition) {
	    TypeDefinition t = (TypeDefinition) o;
	    if (t.type != type)
		return false;
	    if (t.array != array)
		return false;
	    if (isObject() || isEnumValue())
		return (t.value.equals(value));
	    return false;
	}
	else
	    return false;
    }
    
    String getJavaType() {
	if (isEnumValue())
	    return value;
	else if (isArray())
	    return getPrimitiveJavaType() + "[]";
	else if (isPrimitive())
	    return getPrimitiveJavaType();
	else if (isString())
	    return "String";
	else if (isTime_t())
	    return "Date";
	else if (isGpointer())
	    return "Object";
	else if (isGList())
	    return "GListString";
	else
	    return translateNaming();
    }

    String getJNIType() {
	if (isArray())
	    return "jarray";
	if (type == BOOLEAN)
	    return "jboolean";
	if (type == BYTE)
	    return "jbyte";
	if (type == CHAR)
	    return "jchar";
	if (type == DOUBLE)
	    return "jdouble";
	if (type == FLOAT)
	    return "jfloat";
	if (type == INTEGER)
	    return "jint";
	if (type == LONG)
	    return "jlong";
	if (type == SHORT)
	    return "jshort";
	if (type == VOID)
	    return "void";
	if (type == STRING)
	    return "jstring";
	if (type == TIME_T)
	    return "jobject";
	if (type == POINTER)
	    return "jobject";
	if (type == GLIST)
	    return "jobject";
	if (isEnumValue())
	    return "/* " + value + " */ jint";
	else
	    return "jobject";
    }
    
    public String getNativeType() {
	if (isEnumValue())
	    return value;
	if (isArray())
	    return getPrimitiveNativeType() + "*";
	if (isPrimitive())
	    return getPrimitiveNativeType();
	if (isString())
	    return "char *";
	if (isTime_t())
	    return "time_t";
	if (isGpointer())
	    return "gpointer";
	if (isGList())
	    return "GList *";
	// so it's an object
	return value + " *";
    }
    
    String getPrimitiveJavaType() {
	if (type == BOOLEAN)
	    return "boolean";
	if (type == BYTE)
	    return "byte";
	if (type == CHAR)
	    return "char";
	if (type == DOUBLE)
	    return "double";
	if (type == FLOAT)
	    return "float";
	if (type == INTEGER)
	    return "int";
	if (type == LONG)
	    return "long";
	if (type == SHORT)
	    return "shot";
	if (type == STRING)
	    return "String";
	// then type == VOID
	return "void";
    }
    
    String getPrimitiveNativeType() {
	if (type == BOOLEAN)
	    return "int";
	if (type == BYTE)
	    return "short";
	if (type == CHAR)
	    return "char";
	if (type == DOUBLE)
	    return "double";
	if (type == FLOAT)
	    return "float";
	if (type == INTEGER)
	    return "int32_t";
	if (type == LONG)
	    return "int64_t";
	if (type == SHORT)
	    return "short";
	if (type == STRING)
	    return "char *";
	// thus type == VOID
	return "void";
    }
    
    /**
       Returns a String representing the type-encoding, computed as:
       <pre>
       boolean : "Z"
       byte    : "B"
       char    : "C"
       double  : "D"
       float   : "F"
       int     : "I"
       long    : "J"
       short   : "S"
       void    : "V"
       arrays  : "[" + type-encoding of component type
       object  : "L"
       + fully qualified class name with "."'s replaced by "/"'s
       + ";"</pre>
    */
    
    String getPrimitiveSignature() {
	if (type == BOOLEAN)
	    return "Z";
	if (type == BYTE)
	    return "B";
	if (type == CHAR)
	    return "C";
	if (type == DOUBLE)
	    return "D";
	if (type == FLOAT)
	    return "F";
	if (type == INTEGER)
	    return "I";
	if (type == LONG)
	    return "J";
	if (type == SHORT)
	    return "S";
	// then type == VOID
	return "V";
	// FIXME add type for arrays
    }
    
    public String getRawValue() {
	return value;
    }
    
    public String getSignature() {
	if (isEnumValue())
	    return "Ljava.lang.Object"; //;//"I";
	if (isPrimitive())
	    return getPrimitiveSignature();
	if (isArray() && isString())
	    return "[Ljava.lang.String;";
	if (isArray())
	    return '[' + getPrimitiveSignature();
	if (isString())
	    return "Ljava.lang.String;";
	if (isTime_t())
	    return "Ljava.util.Date;";
	if (isGpointer())
	    return "Ljava.lang.Object;";
	if (isGList())
	    return "Lgtk.GListString;";
	// else it's a string
	return new String("L" + makeFullName().replace('.', '/') + ";");
    }
    
    public int getType() {
	return type;
    }
    
    public boolean isArray() {
	return array;
    }
    
    public boolean isEnumValue() {
	return type == ENUM;
    }
    
    public boolean isGList() {
	return type == GLIST;
    }
    
    public boolean isGpointer() {
	return type == POINTER;
    }
    
    public boolean isObject() {
	return type == OBJECT;
    }
    
    public boolean isPrimitive() {
	return ((10 > type) && (type > 0)); // || (type == ENUM));
    }
    
    public boolean isString() {
	return type == STRING;
    }
    
    public boolean isTime_t() {
	return type == TIME_T;
    }
    
    public boolean isVoid() {
	return type == VOID;
    }
    
    public TypeDefinition(int i) {
	if ((11 > i) && (i > 0))
	    type = i;
    }
    
    public String makeFullName() {
	if (translateNaming().startsWith("Gdk")) {
	    return "gdk." + translateNaming();
	}
	else if (translateNaming().startsWith("Gtk")) {
	    return "gtk." + translateNaming();
	}
	else if (translateNaming().startsWith("Gnome")) {
	    return "gnome." + translateNaming();
	}
	else
	    return Parser.currentPackage + "." + translateNaming();
    }
    
    public void setEnumValue(String name) {
	type = ENUM;
	value = name;
    }
    
    public void setEnumValue() {
	type = ENUM;
    }
    
    public String toString() {
	return getType() + ":" + getRawValue() + ":" + translateNaming();
    }
    
    public String translateNaming() {
	if (Parser.naming == Parser.PURE_JAVA) {
	    if (isObject()) {
		for (Enumeration e = Parser.modules.elements(); e.hasMoreElements();) {
		    String mod = (String) e.nextElement();
		    if ((value.toLowerCase()).startsWith(mod.toLowerCase()))
			return value.substring(mod.length());
		}
	    }
	    if (isEnumValue())
		if ((value.toLowerCase()).startsWith(Parser.currentModule.toLowerCase()))
		    return value.substring(Parser.currentModule.length() + 1).toUpperCase();
	    return value;
	}
	return value;
    }
    
    public TypeDefinition() {
    }
}
